#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

from abc import ABCMeta, abstractmethod
from collections import deque

from draugr.writers.mixins.scalar_writer_mixin import ScalarWriterMixin

__all__ = ["Writer", "global_writer", "set_global_writer"]

from typing import Any, Optional, Sequence, MutableMapping


class Writer(ScalarWriterMixin, metaclass=ABCMeta):
    """description"""

    def __enter__(self):
        global GLOBAL_WRITER_STACK, GLOBAL_WRITER
        GLOBAL_WRITER_STACK.appendleft(self)
        GLOBAL_WRITER = self
        return self._open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        global GLOBAL_WRITER, GLOBAL_WRITER_STACK

        if len(GLOBAL_WRITER_STACK) > 0:
            GLOBAL_WRITER_STACK.popleft()  # pop self

        if len(GLOBAL_WRITER_STACK) > 0:
            GLOBAL_WRITER = GLOBAL_WRITER_STACK.popleft()  # then previous
        else:
            GLOBAL_WRITER = None
        return self._close(exc_type, exc_val, exc_tb)

    def close(self) -> Any:
        """description"""
        self._close()

    def open(self) -> Any:
        """description"""
        self._open()

    @abstractmethod
    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        raise NotImplementedError

    @abstractmethod
    def _open(self):
        return self

    def __call__(self, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
        self.scalar(*args, *kwargs)


GLOBAL_WRITER_STACK = deque()
GLOBAL_WRITER = None


def global_writer() -> Optional[Writer]:
    """

    :return:
    :rtype:"""
    global GLOBAL_WRITER
    return GLOBAL_WRITER


def set_global_writer(writer: Writer) -> None:
    """

    :return:
    :rtype:"""
    global GLOBAL_WRITER
    # if GLOBAL_WRITER:
    # GLOBAL_WRITER_STACK TODO: push to stack if existing?

    GLOBAL_WRITER = writer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 7/8/22
           """

__all__ = ["CounterFilter"]

from abc import ABCMeta
from collections import Counter
from typing import Optional, Iterable

from warg import drop_unused_kws, is_none_or_zero_or_negative_or_mod_zero


class CounterFilter(metaclass=ABCMeta):
    """ """

    @drop_unused_kws
    def __init__(
        self,
        *,
        interval: Optional[int] = 1,
        filters: Iterable = None,
        verbose: bool = False
    ):
        """

        :param interval:
        :param filters:
        :param verbose:"""
        self._counter = Counter()

        self._interval = interval
        self.filters = filters
        self._verbose = verbose

    def filter(self, tag: str) -> bool:
        """

            returns a boolean  value, true if to be included, False if to be excluded

            tag is in filter if not None
            and within interval for inclusion

        :param tag:
        :type tag:
        :return:
        :rtype:"""
        is_in_filters = self.filters is None or tag in self.filters
        at_interval = is_none_or_zero_or_negative_or_mod_zero(
            self._interval, self._counter[tag]
        )
        return is_in_filters and at_interval



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

__doc__ = """description"""

__all__ = [
    "COLORS",
    "DECORATIONS",
    "generate_style",
    "sprint",
    "PrintStyle",
    "get_terminal_size",
    "hyperlink_path",
    "hyperlink_url",
]

import os
import shutil
from pathlib import Path
from random import randint
from time import time
from typing import Any, Union

import six

from warg import NOD, passes_kws_to

COLORS = NOD(
    red="31",
    green="32",
    yellow="33",
    # gray='30', #Black,
    blue="34",
    magenta="35",
    cyan="36",
    white="37",
    crimson="38",
)

DECORATIONS = NOD(
    end="0",
    bold="1",
    dim="2",
    italic="3",
    underline="4",
    underline_end="24",  # '4:0',
    double_underline="21",  # '4:2'
    # double_underline_end='24',  # '4:0'
    curly_underline="4:3",
    blink="5",
    reverse_colors="7",
    invisible="8",  # still copyable
    strikethrough="9",
    overline="53",
    hyperlink="8;;",
)


def hyperlink_url(
    link: str, link_id: str = None, style: bool = False, prefix: str = "https://"
) -> str:
    """description"""
    if not link.startswith(prefix):
        if "://" in link:
            link = f'{prefix}{link.split("://")[1]}'
        else:
            link = f"{prefix}{link}"

    if not style:
        return link
    if not link_id:
        link_id = f"{time()}-{randint(0, 999999)}" if link else ""
    return f"\x1b]8;id={link_id};{link}\x1b\\{link}\x1b]8;;\x1b\\"


def hyperlink_path(
    link: Path, link_id: str = None, style: bool = False, linux: bool = True
) -> str:
    """description"""
    if linux:
        link = f"file://{link.resolve()}"
    if not style:
        return link
    if not link_id:
        link_id = f"{time()}-{randint(0, 999999)}" if link else ""
    return f"\x1b]8;id={link_id};{link}\x1b\\{link}\x1b]8;;\x1b\\"


class PrintStyle(object):
    """description"""

    def __init__(self, attributes_joined, end):
        self._attributes_joined = attributes_joined
        self._end = end

    def __call__(self, obj, *args, **kwargs) -> str:
        intermediate_repr = f"\x1b[{self._attributes_joined}m{obj}\x1b[{self._end}m"
        string = six.u(intermediate_repr)
        return string


def generate_style(
    obj: Any = None,
    *,
    color: Union[str, int] = "random",
    bold: bool = False,
    highlight: bool = False,
    underline: bool = False,
    italic: bool = False,
) -> Union[str, PrintStyle]:
    """

    :param obj:
    :type obj:
    :param color:
    :type color:
    :param bold:
    :type bold:
    :param highlight:
    :type highlight:
    :param underline:
    :type underline:
    :param italic:
    :type italic:
    :return:
    :rtype:"""
    attributes = []

    if color == "random":
        num = int(list(COLORS.values())[randint(0, len(COLORS) - 1)])
    elif color in COLORS:
        num = int(COLORS[color])
    elif color in COLORS.values():
        num = int(color)
    elif isinstance(color, int) and color < 39:
        num = color
    else:
        num = int(COLORS["white"])

    if highlight:
        num += 10

    attributes.append(six.u(f"{num}"))

    if bold:
        attributes.append(six.u(f'{DECORATIONS["bold"]}'))

    if underline:
        attributes.append(six.u(f'{DECORATIONS["underline"]}'))

    if italic:
        attributes.append(six.u(f'{DECORATIONS["italic"]}'))

    end = DECORATIONS["end"]

    attributes_joined = six.u(";").join(attributes)

    print_style = PrintStyle(attributes_joined, end)

    if obj:
        return print_style(obj)
    else:
        return print_style


@passes_kws_to(generate_style)
def sprint(obj: Any, print_kws=None, **kwargs) -> None:
    """
    Stylised print. Defaults to stdout
    Valid colors: gray, red, green, yellow, blue, magenta, cyan, white, crimson"""
    if print_kws is None:
        print_kws = {}
    print(generate_style(obj, **kwargs), **print_kws)


def get_terminal_size() -> NOD:
    """

    :return:
    :rtype:"""
    try:
        size = shutil.get_terminal_size()
        columns, rows = size.columns, size.lines
    except:
        rows, columns = (os.getenv("LINES", 25), os.getenv("COLUMNS", 80))

    rows, columns = int(rows), int(columns)

    return NOD(rows=rows, columns=columns)


if __name__ == "__main__":
    sprint(get_terminal_size())
    sprint(hyperlink_path(Path.home()))
    print(hyperlink_path(Path.home()))
    sprint(hyperlink_path(Path.home() / ".xprofile"))
    sprint(hyperlink_url("http://dr.dk"))
    sprint(hyperlink_url("dr.dk"))



#!/usr/bin/env python3
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__project__ = "Draugr"
__author__ = "Christian Heider Nielsen"
__version__ = "1.1.2"
__doc__ = r"""
Created on 27/04/2019

@author: cnheider

"""


try:
    from importlib.resources import files
    from importlib.metadata import PackageNotFoundError
except (ModuleNotFoundError, ImportError) as e:
    from importlib_metadata import PackageNotFoundError
    from importlib_resources import files

from warg import package_is_editable, clean_string, get_version
from apppath import AppPath
from pathlib import Path

# from .drawers import *
# from .writers import *
# from .opencv_utilities import *
# from .torch_utilities import *
# from .stopping import *
# from .numpy_utilities import *
# from .visualisation import *
# from .metrics import *
# from .python_utilities import *

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

# with open(Path(__file__).parent.parent / "README.md", "r") as this_init_file:
#    __doc__ += this_init_file.read()

__all__ = [
    "PROJECT_APP_PATH",
    "PROJECT_NAME",
    "PROJECT_VERSION",
    "PROJECT_ORGANISATION",
    "PROJECT_AUTHOR",
    "PROJECT_YEAR",
    "INCLUDE_PROJECT_READMES",
    "PACKAGE_DATA_PATH",
]

PROJECT_ORGANISATION = clean_string("Pything")
PROJECT_NAME = clean_string(__project__)
PROJECT_VERSION = __version__
PROJECT_YEAR = 2018
PROJECT_AUTHOR = clean_string(__author__)
PROJECT_APP_PATH = AppPath(app_name=PROJECT_NAME, app_author=PROJECT_AUTHOR)
INCLUDE_PROJECT_READMES = False

__url__ = f"https://github.com/{PROJECT_ORGANISATION.lower()}/{PROJECT_NAME}"

PACKAGE_DATA_PATH = files(PROJECT_NAME) / "data"

try:
    DEVELOP = package_is_editable(PROJECT_NAME)
except PackageNotFoundError as e:
    DEVELOP = True


__version__ = get_version(__version__, append_time=DEVELOP)

__version_info__ = tuple(int(segment) for segment in __version__.split("."))

if __name__ == "__main__":
    print(__version__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/22/22
           """

__all__ = ["VariationalAutoEncoder"]

from abc import abstractmethod

import torch
from draugr.torch_utilities.tensors.to_tensor import to_tensor


class VariationalAutoEncoder(torch.nn.Module):
    """
    General purpose variational auto encoder
    """

    def __init__(self, latent_size: int = 10):
        super().__init__()
        self._latent_size = latent_size

    @abstractmethod
    def encode(self, *x: torch.Tensor) -> torch.Tensor:
        """

        Args:
          *x:
        """
        raise NotImplementedError

    @abstractmethod
    def decode(self, *x: torch.Tensor) -> torch.Tensor:
        """

        Args:
          *x:
        """
        raise NotImplementedError

    def sample(self, *x, num=1) -> torch.Tensor:
        """

        :param x:
        :type x:
        :param num:
        :type num:
        :return:
        :rtype:"""
        return self.decode(
            # z
            torch.randn(num, self._latent_size).to(
                device=next(self.parameters()).device
            ),
            *x,
        ).to("cpu")

    @staticmethod
    def reparameterise(mean, log_var) -> torch.Tensor:
        """
        TODO: MOVE TO SUB ANOTHER CLASS, this is for a guassian specialisation of VAE

                reparameterisation trick

                :param mean:
                :type mean:
                :param log_var:
                :type log_var:
                :return:
                :rtype:"""
        std = log_var.div(2).exp()  # e^(1/2 * log(std^2))
        eps = torch.randn_like(std)  # random ~ N(0, 1)
        return eps.mul(std).add_(mean)  # Reparameterise distribution

    def sample_from(self, *encoding) -> torch.Tensor:
        """

        Args:
          *encoding:

        Returns:

        """
        sample = to_tensor(*encoding).to(device=next(self.parameters()).device)
        assert sample.shape[-1] == self._latent_size, (
            f"sample.shape[-1]:{sample.shape[-1]} !="
            f" self._encoding_size:{self._latent_size}"
        )
        return self.decode(*sample).to("cpu")



import contextlib
from typing import Callable

from warg import sink, Number

__all__ = ["ImprovementDetector", "OverfitDetector"]


class ImprovementDetector(contextlib.AbstractContextManager):
    """description"""

    def __init__(
        self,
        patience: int,
        writer: Callable = print,
        minimization: bool = True,
        callback: Callable = None,
    ):
        """
        NOTE: strictly greater or less than is considered as improvement

        :param patience:
        :type patience:
        :param writer:
        :type writer:
        :param minimization:
        :type minimization:
        """
        self._patience = patience
        self._writer = writer
        self._count = 0
        self._minimization = minimization  # as opposed to maximization
        self._best_value = None
        self._callback = (lambda: True) if callback is None else callback
        self._best_idx = None

    def __call__(self, value: Number) -> bool:
        if self._best_value is None:
            self._best_value = value
            return True

        if self._minimization:
            if value >= self._best_value:
                self._count += 1
                if self._verbose:
                    self._writer(
                        f"No improvement since last update: {value}>{self._best_value}"
                    )
            else:
                self._count = 0
                self._best_value = value
        else:
            if value <= self._best_value:
                self._count += 1
            else:
                self._count = 0
                self._best_value = value

        if self._count >= self._patience:
            self._writer(f"No improvement detected, patience reached")
            return False
        else:
            return self._callback()

    def reset(self) -> None:
        """description"""
        self._count = 0
        self._best_value = None
        self._writer(f"Improvement detector reset")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return True


class OverfitDetector(contextlib.AbstractContextManager):
    """description"""

    def __init__(
        self,
        patience: int,
        writer: Callable = print,
        minimization: bool = True,
        callback: Callable = None,
        verbose: bool = False,
    ):
        """
        NOTE: equality, greater or less than

        :param patience:
        :type patience:
        :param writer:
        :type writer:
        :param minimization:
        :type minimization:
        """
        self._patience = patience
        self._writer = writer
        self._count = 0
        self._minimization = minimization  # as opposed to maximization
        self._best_value = None
        self._callback = sink if callback is None else callback
        self._verbose = verbose

    def __call__(self, value: Number) -> bool:
        if self._best_value is None:
            self._best_value = value
            return False

        if self._minimization:
            if value > self._best_value:
                self._count += 1
                if self._verbose:
                    self._writer(f"Worse than last update: {value}>{self._best_value}")
            else:
                self._count = 0
                self._best_value = value
        else:
            if value < self._best_value:
                self._count += 1
                if self._verbose:
                    self._writer(f"Worse than last update: {value}<{self._best_value}")
            else:
                self._count = 0
                self._best_value = value

        if self._count >= self._patience:
            self._writer(
                f"Overfit detected, patience reached {value}>{self._best_value} with patience {self._patience}"
            )
            return True
        else:
            self._callback()
            return False

    def reset(self) -> None:
        """description"""
        self._count = 0
        self._best_value = None
        self._writer(f"Overfit detector reset")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return True


if __name__ == "__main__":
    with OverfitDetector(patience=3, writer=print) as is_overfitting:
        for i in range(10):
            is_overfitting(i)
        is_overfitting.reset()
        for i in range(10):
            is_overfitting(-i)

        is_overfitting.reset()
        print("start training")
        for i in range(2):
            is_overfitting(i)
        is_overfitting(0)
        print("start overfitting")
        for i in range(10):
            print(is_overfitting(i), i)

    print("\n\n")

    with ImprovementDetector(patience=3, writer=print) as is_improving:
        for i in range(10):
            is_improving(i)
        is_improving.reset()
        for i in range(10):
            is_improving(-i)
        is_improving.reset()
        print("start training")
        for i in range(2):
            is_improving(i)
        is_improving(0)
        print("start not improving")
        for i in range(10):
            print(is_improving(i), i)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-12-2020
           """

from warg import ensure_in_sys_path, find_nearest_ancestral_relative

ensure_in_sys_path(find_nearest_ancestral_relative("draugr").parent)
from draugr.numpy_utilities import root_mean_square


def test_root_mean_square_signed(asad: int = 60):
    s = [i - asad // 2 for i in range(asad)]
    a = root_mean_square(s)
    # assert a


def test_root_mean_square_unsigned(asad: int = 60):
    s = [i + asad // 2 for i in range(asad)]
    a = root_mean_square(s)
    # assert a == mean(s)..



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/2/22
           
           """

__all__ = ["choose_progress_bar"]

from itertools import count
from typing import Iterator

from draugr.python_utilities.platform_context import in_ipynb
from draugr.visualisation.progress.eta_bar import ETABar
from draugr.visualisation.progress.progress_bar import progress_bar

from warg import LambdaContext


class IteratorWrapper(Iterator):
    """
    #TO be removed
    """

    def __next__(self):
        return getattr(self.callable_, self.entered)()

    def __init__(self, iter_, cal_):
        self.iter_ = iter_
        self.callable_ = cal_

    def __iter__(self):
        a = iter(self.entered)
        for c in count():
            yield getattr(self.callable_, a)()

    def __enter__(self):
        self.entered = self.iter_.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.entered.__exit__(exc_type, exc_val, exc_tb)


def choose_progress_bar(*args, total=None, verbose: bool = False, **kwargs) -> callable:
    """
    #TODO: REwrite this!

    :param args:
    :type args:
    :param total:
    :type total:
    :param verbose:
    :type verbose:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """
    if in_ipynb(verbose=verbose):
        print("*args where dropped")  # TODO: POOPY, dont do this
        return ETABar(max=total, verbose=verbose, **kwargs)
    return LambdaContext(progress_bar(*args, total=total, verbose=verbose, **kwargs))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/07/2020
           """

__all__ = [
    "load_model_parameters",
    "load_latest_model_parameters",
    "save_parameters_and_configuration",
    "save_model_parameters",
]

import datetime
import os
from pathlib import Path
from typing import Optional, Tuple, Union

import torch
from draugr.torch_utilities.persistence.config import (
    ensure_directory_exist,
    save_config,
)
from torch import nn
from torch.nn.modules.module import Module
from torch.optim import Optimizer

from warg import drop_unused_kws, passes_kws_to

PARAMETER_EXTENSION = ".parameters"
CONFIG_EXTENSION = ".py"
OPTIMISER_EXTENSION = ".optimiser"


@drop_unused_kws
def load_latest_model_parameters(
    model: Union[torch.nn.Module, nn.Parameter],
    *,
    optimiser: Optimizer = None,
    model_name: str,
    model_directory: Path,
) -> Tuple[Union[torch.nn.Module, Tuple[torch.nn.Module, Optimizer]], bool]:
    """

    inplace but returns model

    :param optimiser:
    :param model:
    :type model:
    :param model_directory:
    :param model_name:
    :return:"""
    model_loaded = False
    optimiser_loaded = False
    model_path = model_directory / model_name

    list_of_files = list(model_path.glob(f"*{PARAMETER_EXTENSION}"))
    if len(list_of_files) == 0:
        print(
            f"Found no previous models with extension {PARAMETER_EXTENSION} in {model_path}"
        )
    else:
        latest_model_parameter_file = max(list_of_files, key=os.path.getctime)
        print(f"loading previous model parameters: {latest_model_parameter_file}")

    if isinstance(model, torch.nn.Module):
        model.load_state_dict(torch.load(str(latest_model_parameter_file)))
        model_loaded = True
    elif isinstance(model, nn.Parameter):
        model.data = torch.load(str(latest_model_parameter_file))
        model_loaded = True
    else:
        # print(f"model must be a torch.nn.Module or nn.Parameter")
        raise TypeError(f"model must be a torch.nn.Module or nn.Parameter")

    if optimiser:
        opt_st_d_file = latest_model_parameter_file.with_suffix(OPTIMISER_EXTENSION)
        if opt_st_d_file.exists():
            optimiser.load_state_dict(torch.load(str(opt_st_d_file)))
            print(f"loading previous optimiser state: {opt_st_d_file}")
            optimiser_loaded = True

        return (model, optimiser), (model_loaded, optimiser_loaded)
    return model, model_loaded


load_model_parameters = load_latest_model_parameters


@passes_kws_to(save_config)
def save_parameters_and_configuration(
    *,
    model: Union[torch.nn.Module, nn.Parameter],
    model_save_path: Path,
    optimiser: Optional[Optimizer] = None,
    optimiser_save_path: Optional[Path] = None,
    config_save_path: Optional[Path] = None,
    loaded_config_file_path: Optional[Path] = None,
) -> None:
    """

    save model parameters and configuration to disk

    :param optimiser:
    :type optimiser:
    :param optimiser_save_path:
    :type optimiser_save_path:
    :param model:
    :param model_save_path:
    :param config_save_path:
    :param loaded_config_file_path:
    :return:"""
    if isinstance(model, torch.nn.Module):
        torch.save(model.state_dict(), str(model_save_path))
    elif isinstance(model, nn.Parameter):
        torch.save(model.data, str(model_save_path))
    else:
        raise TypeError(f"model must be a torch.nn.Module or nn.Parameter")
    if optimiser:
        torch.save(optimiser.state_dict(), str(optimiser_save_path))
    if loaded_config_file_path:
        save_config(config_save_path, loaded_config_file_path)


@drop_unused_kws
def save_model_parameters(
    model: Module,
    *,
    model_name: str,
    save_directory: Path,
    optimiser: Optional[Optimizer] = None,
    config_file_path: Optional[Path] = None,
    verbose: bool = False,
) -> None:
    """

    save model parameters and optionally configuration file and optimiser state dict

    :param verbose:
    :type verbose:
    :param optimiser:
    :param model:
    :param save_directory:
    :param config_file_path:
    :param model_name:
    :return:"""
    model_date = datetime.datetime.now()

    model_time_rep = model_date.strftime("%Y%m%d%H%M%S")
    model_save_path = save_directory / model_name / f"{model_time_rep}"
    ensure_directory_exist(model_save_path.parent)

    saved = False
    try:
        save_parameters_and_configuration(
            model=model,
            model_save_path=model_save_path.with_suffix(PARAMETER_EXTENSION),
            optimiser=optimiser,
            optimiser_save_path=(
                model_save_path.parent / f"{model_time_rep}"
            ).with_suffix(OPTIMISER_EXTENSION),
            loaded_config_file_path=config_file_path,
            config_save_path=(model_save_path.parent / f"{model_time_rep}").with_suffix(
                CONFIG_EXTENSION
            ),
        )
        saved = True
    except FileNotFoundError as e:
        print(e)
        while not saved:
            model_save_path = (
                Path(input("Enter another file path: ")).expanduser().resolve()
            )
            ensure_directory_exist(model_save_path.parent)
            try:
                save_parameters_and_configuration(
                    model=model,
                    model_save_path=model_save_path.endswith(PARAMETER_EXTENSION),
                    optimiser=optimiser,
                    optimiser_save_path=(
                        model_save_path.parent / f"{model_time_rep}"
                    ).with_suffix(OPTIMISER_EXTENSION),
                    loaded_config_file_path=config_file_path,
                    config_save_path=(
                        model_save_path.parent / f"{model_time_rep}"
                    ).with_suffix(CONFIG_EXTENSION),
                )
                saved = True
            except FileNotFoundError as e:
                print(e)
                saved = False
    if verbose:
        if saved:
            print(
                f"Successfully saved model parameters, optimiser state and configuration at names {[model_save_path.with_suffix(a) for a in (PARAMETER_EXTENSION, OPTIMISER_EXTENSION, CONFIG_EXTENSION)]}"
            )
        else:
            print(f"Was unsuccessful at saving model or configuration")



__author__ = "Christian Heider Nielsen"
__doc__ = r"""description"""
__all__ = [
    "SIMPLE_TIMER_TEMPLATE",
]

# OnCalendar=*-*-* 06:00:00
# with the same unit prefix name
SIMPLE_TIMER_TEMPLATE = """
[Unit]
Description=Run {service} {interval}

[Timer]
OnCalendar={on_calendar}
Persistent=true

[Install]
WantedBy=timers.target
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Union, Sequence

import torch

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["GraphWriterMixin"]


class GraphWriterMixin(ABC):
    """
    Writer mixin that provides an interface for 'writing' graphs"""

    @abstractmethod
    def graph(
        self,
        model: torch.nn.Module,
        input_to_model: Union[torch.Tensor, Sequence[torch.Tensor]],
        **kwargs
    ) -> None:
        """

        Write graph


            :param model:
            :param input_to_model:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



from abc import ABC, abstractmethod
from itertools import cycle
from typing import MutableMapping

from draugr.python_utilities import CounterFilter

from warg import Number, passes_kws_to, drop_unused_kws

__all__ = ["ScalarWriterMixin"]


class ScalarWriterMixin(CounterFilter, ABC):
    """description"""

    @abstractmethod
    def _scalar(self, tag: str, value: float, step: int):
        raise NotImplementedError

    @passes_kws_to(CounterFilter.__init__)
    @drop_unused_kws
    def __init__(self, **kwargs: MutableMapping):
        super().__init__(**kwargs)
        self._blip_iterators = {}

    def scalar(self, tag: str, value: Number, step_i: int = None) -> None:
        """

        :param tag:
        :type tag:
        :param value:
        :type value:
        :param step_i:
        :type step_i:"""
        if step_i:
            self._counter[tag] = step_i
        else:
            self._counter[tag] += 1

        if self.filter(tag):
            self._scalar(tag, value, self._counter[tag])

    def blip(self, tag: str, step_i: int = None) -> None:
        """

        :param tag:
        :type tag:
        :param step_i:
        :type step_i:"""
        if tag not in self._blip_iterators:
            self._blip_iterators[tag] = iter(cycle(range(2)))
        self.scalar(tag, next(self._blip_iterators[tag]), step_i=step_i)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/10/2019
           """

__all__ = ["set_lr", "exponential_lr_decay", "average_learning_rate"]

from statistics import mean

import torch
from torch.optim.optimizer import Optimizer


def set_lr(optimizer: torch.optim.Optimizer, lr: float) -> None:
    """

    :param optimizer:
    :type optimizer:
    :param lr:
    :type lr:"""
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def average_learning_rate(optimizer: torch.optim.Optimizer) -> float:
    """

    :param optimizer:
    :type optimizer:
    :return:
    :rtype:
    """
    learning_rates = []
    for param_group in optimizer.param_groups:
        learning_rates.append(param_group["lr"])
    return mean(learning_rates)


def exponential_lr_decay(
    optim: Optimizer,
    *,
    initial_learning_rate: float,
    step: int,
    decay_rate: float,
    decay_steps: float,
) -> None:
    """
    This method will decay the learning rate continuously for all
        parameter groups in the optimizer by the function:

        lr = lr_0 * decay^(step / decay_steps)

    :param optim: The optimiser to modify
    :type optim: Optimizer
    :param initial_learning_rate: The initial learning rate (lr_0)
    :type initial_learning_rate: float
    :param step: The current step in training
    :type step: int
    :param decay_rate: The rate at which to decay the learning rate
    :type decay_rate: float
    :param decay_steps: The number of steps before the learning rate is
                             applied in full.
    :type decay_steps: float
    :return:
    :rtype:
    """

    decay_rate = decay_rate ** (step / decay_steps)
    lr = initial_learning_rate * decay_rate
    for group in optim.param_groups:
        group["lr"] = lr


if __name__ == "__main__":

    def _main():
        from torch.nn import Parameter

        s = (Parameter(torch.ones(1, 2)),)
        from torch.optim import Adam

        optim = Adam(s)
        print(average_learning_rate(optim))

    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 8/30/22
           """

__all__ = []

from warg.os_utilities.os_platform import has_x_server


def windows_display_test():
    import subprocess, os

    if os.name == "nt":  # Windows
        import wmi

        try:
            wmi.WMI().computer.Win32_VideoController()[0]  # Tested on Windows 10
            return 1
        except:
            pass

    elif os.name == "posix":  # Linux
        out = subprocess.getoutput(
            "sudo lshw -c video | grep configuration"
        )  # Tested on CENTOS 7 and Ubuntu
        if out:
            return 1


if __name__ == "__main__":
    print(has_x_server())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

import subprocess
from enum import Enum
from pathlib import Path
from typing import Any, Sequence, Union

import numpy
from cycler import Cycler
from draugr.visualisation.matplotlib_utilities.quirks import auto_post_hatch
from draugr.visualisation.matplotlib_utilities.styles.annotation import (
    rt_ann_transform,
    semi_opaque_round_tight_bbox,
)
from draugr.visualisation.matplotlib_utilities.styles.cyclers import (
    monochrome_hatch_cycler,
    monochrome_line_cycler,
    simple_hatch_cycler,
)
from matplotlib import patches, pyplot, rcParams
from matplotlib.legend_handler import HandlerErrorbar
from sorcery import assigned_names

__all__ = [
    "denormalise_minusoneone",
    "matplotlib_bounding_box",
    "remove_decoration",
    "use_monochrome_style",
    "decolorise_plot",
    "save_embed_fig",
    "latex_clean_label",
    "make_errorbar_legend",
    "annotate_point",
    "MatplotlibHorizontalAlignment",
    "MatplotlibVerticalAlignment",
]

from matplotlib.patches import Rectangle
from matplotlib.pyplot import legend

from warg import Number, passes_kws_to

# Cycler
##Addition
###Equal length Cycler s with different keys can be added to get the ‘inner’ product of two cycles
##Multiplication
###which gives the ‘outer product’ of the two cycles (same as itertools.prod() )

from matplotlib.axes import Axes, ErrorbarContainer


class MatplotlibHorizontalAlignment(Enum):
    """
    Horizontal alignment of text in matplotlib

    """

    (center, right, left) = assigned_names()


class MatplotlibVerticalAlignment(Enum):
    """
    Vertical alignment of text in matplotlib

    """

    (center, top, bottom, baseline, center_baseline) = assigned_names()


def latex_clean_label(s: str) -> str:
    """Clean label for troublesome symbols"""
    return s.replace("_", " ")  # .replace('\\',' ').replace('/',' ')


def make_errorbar_legend(ax: Axes = None) -> None:
    """size adjusted legend for error bars, default does not work well with different linestyles"""

    if ax is None:
        ax = pyplot.gca()
    ax.legend(
        handlelength=5,
        handleheight=3,
        handler_map={ErrorbarContainer: HandlerErrorbar(xerr_size=0.9, yerr_size=0.9)},
    )


def annotate_point(ax: Axes, x: Sequence, y: Sequence, t: Any) -> None:
    """

    Annotate point with text

    :param ax:
    :param x:
    :param y:
    :param t:
    :return:
    """
    if ax is None:
        ax = pyplot.gca()
    ax.annotate(
        f"{t:.2f}",
        (x, y),
        textcoords="offset points",
        fontsize="xx-small",
        bbox=semi_opaque_round_tight_bbox,
        annotation_clip=True,
        # see details https://github.com/matplotlib/matplotlib/issues/14354#issuecomment-523630316
        clip_on=True,
        **rt_ann_transform,
    )


@passes_kws_to(pyplot.savefig)
def save_embed_fig(
    path: Union[Path, str],
    bbox_inches: Union[Number, str] = "tight",
    transparent: bool = True,
    attempt_fix_empty_white_space: bool = False,
    post_process_crop: bool = False,
    ax: Axes = None,
    suffix: str = ".pdf",
    **kwargs,
) -> None:
    """Save fig for latex pdf embedding"""

    if attempt_fix_empty_white_space:  # remove it
        # pyplot.gca().set_axis_off()
        # pyplot.axis('off') # this rows the rectangular frame
        # ax.get_xaxis().set_visible(False) # this removes the ticks and numbers for x axis
        # ax.get_yaxis().set_visible(False) # this removes the ticks and numbers for y axis
        if ax is None:
            ax = pyplot.gca()
        pyplot.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        pyplot.margins(0, 0)
        ax.xaxis.set_major_locator(pyplot.NullLocator())
        ax.yaxis.set_major_locator(pyplot.NullLocator())

    """
clip_box = Bbox(((0,0),(300,300)))
for o in pyplot.findobj():
o.set_clip_on(True)
o.set_clip_box(clip_box)

"""

    if not isinstance(path, Path):
        path = Path(path)

    path_str = str(path.with_suffix(suffix))
    pyplot.savefig(
        path_str,
        bbox_inches=bbox_inches,
        transparent=transparent,
        **kwargs,
    )

    if (
        post_process_crop and suffix == ".pdf"
    ):  # Generally a good idea since matplotlib does not exclude invisible parts(eg. data points or anchors) of the plot.
        from pdfCropMargins import crop  # pip install pdfCropMargins

        crop(
            [
                "-p",
                "0",  # remove all percentage of old margin
                "-a",
                "-5",  # add 5bp margin around all
                path_str,
            ]
        )


def remove_decoration(ax: Axes) -> None:
    """
    removes decoration from axes

    :param ax:
    :type ax:
    :return:
    :rtype:
    """
    transparent = (1.0, 1.0, 1.0, 0.0)

    ax.w_xaxis.set_pane_color(transparent)
    ax.w_yaxis.set_pane_color(transparent)
    ax.w_zaxis.set_pane_color(transparent)

    ax.w_xaxis.line.set_color(transparent)
    ax.w_yaxis.line.set_color(transparent)
    ax.w_zaxis.line.set_color(transparent)

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])


def denormalise_minusoneone(
    t: Union[numpy.ndarray, Number], coordinates: Union[numpy.ndarray, Number]
) -> Union[numpy.ndarray, Number]:
    """

    #denormalise from -1 to 1 to 0 to 1

    :param t:
    :type t:
    :param coordinates:
    :type coordinates:
    :return:
    :rtype:
    """
    return 0.5 * ((coordinates + 1.0) * t)


def decolorise_plot(ax_: Axes, inverted: bool = False) -> None:
    """

    set black and white edge colors and face colors respectively. Converse if inverted.
    """
    edge_color = "k"
    face_color = "w"
    if inverted:
        face_color, edge_color = edge_color, face_color
    pyplot.setp(ax_.artists, edgecolor=edge_color, facecolor=face_color)
    pyplot.setp(ax_.lines, color=edge_color)


def matplotlib_bounding_box(
    x: Number, y: Number, size: Number, color: str = "w"
) -> Rectangle:
    """

    Create a bounding box in matplotlib

    :param x:
    :type x:
    :param y:
    :type y:
    :param size:
    :type size:
    :param color:
    :type color:
    :return:
    :rtype:
    """
    x = int(x - (size / 2))
    y = int(y - (size / 2))
    rect = patches.Rectangle(
        (x, y), size, size, linewidth=1, edgecolor=color, fill=False
    )
    return rect


def use_monochrome_style(
    prop_cycler: Cycler = monochrome_line_cycler,
    # ONLY COLOR AND LINESTYLE MAKES SENSE FOR NOW, matplotlib seems very undone in this api atleast for bars
) -> None:
    """

    Set matplotlib style to monochrome

    :param prop_cycler:
    """
    # from matplotlib.pyplot import axes, grid

    # from matplotlib import lines, markers
    # print(lines.lineStyles.keys(),markers.MarkerStyle.markers.keys())

    # print(rcParams.keys())

    # pyplot.style.use('default')
    # print(pyplot.style.available)
    # print(matplotlib.matplotlib_fname())

    pyplot.style.use(Path(__file__).parent / "styles" / "monochrome.mplstyle")
    if pyplot.rcParams["text.usetex"]:
        try:
            report = subprocess.check_output("latex -v", stderr=subprocess.STDOUT)
        except FileNotFoundError as exc:
            msg = f'No tex: {"latex"}'
            # raise RuntimeError(msg)
            print(f"{msg}, disabling")
            pyplot.rcParams.update({"text.usetex": False})
    rcParams.update({"axes.prop_cycle": prop_cycler})
    # auto_post_print_dpi()


if __name__ == "__main__":

    def asiuhda() -> None:
        """
        :rtype: None
        """
        use_monochrome_style()
        bar_styles = monochrome_hatch_cycler()
        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.bar(x, numpy.random.randint(2, 10), **next(bar_styles), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        pyplot.show()

    def asiuh214da() -> None:
        """
        :rtype: None
        """
        pyplot.style.use(Path(__file__).parent / "styles" / "monochrome.mplstyle")
        line_styles = monochrome_line_cycler()
        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.plot(numpy.random.rand(10), **next(line_styles), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        pyplot.show()

    def asiuasdashda() -> None:
        """
        :rtype: None
        """
        use_monochrome_style(prop_cycler=monochrome_hatch_cycler)

        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.bar(x, numpy.random.randint(2, 10), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        auto_post_hatch(ax, simple_hatch_cycler)
        pyplot.show()

    def asiuhdsada_non() -> None:
        """
        :rtype: None
        """
        fig, ax = pyplot.subplots(1, 1)
        for x in range(3):
            ax.plot(numpy.random.rand(10), label=f"{x}")

        legend()
        pyplot.show()

    def asiuhd21412sada() -> None:
        """
        :rtype: None
        """
        use_monochrome_style()
        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.plot(numpy.random.rand(10), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        pyplot.show()

    # asiuhda()
    # asiuasdashda()
    asiuhd21412sada()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/30/22
           """

__all__ = ["laplace", "divergence", "gradient", "hessian", "jacobian"]

from typing import Tuple

import torch
from torch.autograd import grad


def laplace(y: torch.Tensor, x: torch.Tensor) -> float:
    """
    laplacian of y wrt x

    y: shape (meta_batch_size, num_observations, channels)
    x: shape (meta_batch_size, num_observations, 2)

    :param y:
    :type y:
    :param x:
    :type x:
    :return:
    :rtype:
    """
    return divergence(gradient(y, x), x)


def divergence(y: torch.Tensor, x: torch.Tensor) -> float:
    """
    divergence of y wrt x

    y: shape (meta_batch_size, num_observations, channels)
    x: shape (meta_batch_size, num_observations, 2)

    :param y:
    :type y:
    :param x:
    :type x:
    :return:
    :rtype:
    """
    div = 0.0
    for i in range(y.shape[-1]):
        div += torch.autograd.grad(
            y[..., i], x, torch.ones_like(y[..., i]), create_graph=True
        )[0][..., i : i + 1]
    return div


def gradient(
    y: torch.Tensor, x: torch.Tensor, grad_outputs: torch.Tensor = None
) -> torch.Tensor:
    """
    gradient of y wrt x

    y: shape (meta_batch_size, num_observations, channels)
    x: shape (meta_batch_size, num_observations, 2)

    :param y:
    :type y:
    :param x:
    :type x:
    :param grad_outputs:
    :type grad_outputs:
    :return:
    :rtype:
    """
    if grad_outputs is None:
        grad_outputs = torch.ones_like(y)
    return torch.autograd.grad(y, [x], grad_outputs=grad_outputs, create_graph=True)[0]


def hessian(y: torch.Tensor, x: torch.Tensor) -> Tuple[torch.Tensor, int]:
    """hessian of y wrt x
    y: shape (meta_batch_size, num_observations, channels)
    x: shape (meta_batch_size, num_observations, 2)
    """
    meta_batch_size, num_observations = y.shape[:2]
    grad_y = torch.ones_like(y[..., 0]).to(y.device)
    h = torch.zeros(
        meta_batch_size, num_observations, y.shape[-1], x.shape[-1], x.shape[-1]
    ).to(y.device)
    for i in range(y.shape[-1]):
        # calculate dydx over batches for each feature value of y
        dydx = grad(y[..., i], x, grad_y, create_graph=True)[0]

        # calculate hessian on y for each x value
        for j in range(x.shape[-1]):
            h[..., i, j, :] = grad(dydx[..., j], x, grad_y, create_graph=True)[0][
                ..., :
            ]

    status = 0
    if torch.any(torch.isnan(h)):
        status = -1
    return h, status


def jacobian(y: torch.Tensor, x: torch.Tensor) -> Tuple[torch.Tensor, int]:
    """jacobian of y wrt x"""
    meta_batch_size, num_observations = y.shape[:2]
    jac = torch.zeros(meta_batch_size, num_observations, y.shape[-1], x.shape[-1]).to(
        y.device
    )  # (meta_batch_size*num_points, 2, 2)
    for i in range(y.shape[-1]):
        # calculate dydx over batches for each feature value of y
        y_flat = y[..., i].view(-1, 1)
        jac[:, :, i, :] = grad(y_flat, x, torch.ones_like(y_flat), create_graph=True)[0]

    status = 0
    if torch.any(torch.isnan(jac)):
        status = -1

    return jac, status



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ = this_init_file.read()

try:
    from .system import *
    from .datasets import *
    from .distributions import *
    from .generators import *
    from .images import *
    from .operations import *
    from .optimisation import *
    from .persistence import *
    from .tensors import *
    from .writers import *
    from .sessions import *
    from .exporting import *
    from .evaluation import *
    from .opencv import *
    from .architectures import *
except ImportError as ix:
    this_package_name = Path(__file__).parent.name
    this_package_reqs = (
        Path(__file__).parent.parent.parent
        / "requirements"
        / f"requirements_{this_package_name}.txt"
    )
    if this_package_reqs.exists():
        print(
            f"Make sure requirements is installed for {this_package_name}, see {this_package_reqs}"
        )  # TODO: PARSE WHAT is missing and print
    raise ix

if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-09-2020
           """
__all__ = ["get_screen_resolution"]

import pygame


def get_screen_resolution():
    """

    :return:
    :rtype:
    """
    info = pygame.display.Info()
    width = info.current_w
    height = info.current_h
    return width, height



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 8/24/22
           """

__all__ = [
    "in_ipynb",
    "in_docker",
    "can_ping_subprocess",
    "can_ping",
    "hostname_resolves",
]

import sys
import os
from warg.os_utilities import is_windows
from warg.functions import text_in_file
from pathlib import Path
import subprocess
import platform


def in_ipynb(verbose: bool = False) -> bool:
    """

    :return:
    :rtype:
    """
    try:
        from IPython import get_ipython
        import jupyter

        shell = get_ipython().__class__.__name__
        if verbose:
            print(f"found shell: {shell}")
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        if verbose:
            print(f"Probably standard Python interpreter")
        return False  # Probably standard Python interpreter
    except ModuleNotFoundError:
        if "ipykernel" in sys.modules:
            if verbose:
                print(f"Found ipykernel in sys.modules")
            return True

        if verbose:
            print(f"Did not find Ipython")
        return False  # Did not find Ipython


def in_docker(
    # hostname: str = "host.docker.internal"
) -> bool:
    """ """
    return (
        Path("/.dockerenv").exists()  # short circuits
        or text_in_file("docker", Path("/proc/self/cgroup"))  # short circuits
        # or (hostname_resolves(hostname) and can_ping_subprocess(hostname))
    )


def can_ping(hostname: str) -> bool:
    return not bool(os.system(f"ping -{'n' if is_windows() else 'c'} 1 {hostname}"))


def hostname_resolves(hostname: str) -> bool:
    try:
        import socket

        socket.gethostbyname(hostname)
        return True
    except socket.error:
        return False


def can_ping_subprocess(hostname: str) -> bool:
    try:
        return not bool(
            subprocess.check_output(
                f"ping -{'n' if is_windows() else 'c'} 1 {hostname}",
                shell=True,
            )
        )
    except Exception:
        return False


if __name__ == "__main__":
    print(in_ipynb())
    print(in_docker())
    # print(hostname_resolves("host.docker.internal"))
    print(can_ping_subprocess("host.docker.internal"))
    print(can_ping("host.docker.internal"))
    # print(hostname_resolves())



from enum import Enum
from typing import Iterable

from sorcery import assigned_names

__all__ = ["ExtensionEnum", "match_return_code"]

ESC_CHAR = chr(27)


def match_return_code(ret_val, chars: Iterable[str] = ("q", ESC_CHAR)) -> bool:
    """

    :param ret_val:
    :type ret_val:
    :param chars:
    :type chars:
    :return:
    :rtype:
    """
    if ret_val:
        return any(ret_val & 0xFF == ord(c) for c in chars)
    return False


class ExtensionEnum(Enum):
    png, exr = assigned_names()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 8/10/22
           """

__all__ = ["BidirectionalIterator", "prev"]

from typing import Iterator, Any


class BidirectionalIterator(Iterator):
    """
    Enable iteration forward and backward

    """

    def __next__(self) -> Any:
        self.i += 1
        if self.i < len(self.history):
            return self.history[self.i]
        else:
            elem = next(self.iterator)
            self.history.append(elem)
            return elem

    def __init__(self, iterator: Iterator):
        self.iterator = iterator
        self.history = [
            None,
        ]
        self.i = 0

    def next(self):
        """

        :return:
        :rtype:
        """
        return self.__next__()

    def prev(self):
        """

        :return:
        :rtype:
        """
        self.i -= 1
        if self.i == 0:
            raise StopIteration
        else:
            return self.history[self.i]


def prev(iterator: BidirectionalIterator) -> Any:
    """

    :param iterator:
    :type iterator:
    :return:
    :rtype:
    """
    return iterator.prev()


if __name__ == "__main__":

    def asdasf() -> None:
        a = BidirectionalIterator(iter([1, 2, 3, 4, 5, 6]))
        print(next(a))
        print(next(a))
        print(prev(a))
        print(next(a))
        print(a.next())
        print(a.prev())
        print(a.prev())
        print(a.next())
        print(a.next())

    asdasf()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 8/30/22
           """

__all__ = ["Matrix"]


class Matrix(list):
    def __matmul__(self, B):  # Matrix multiplication A @ B
        A = self

        N = len(A)
        M = len(A[0])
        P = len(B[0])

        result = []
        for i in range(N):
            row = [0] * P
            result.append(row)

        for i in range(N):
            for j in range(P):
                for k in range(M):
                    result[i][j] += A[i][k] * B[k][j]
        return result


if __name__ == "__main__":

    def hasdhuh():
        # Example
        A = Matrix([[2, 0], [1, 9]])
        B = Matrix([[3, 9], [4, 7]])
        print(A @ B)

    hasdhuh()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 10/3/22
           """

__all__ = ["Sine", "Cosine"]

import torch
from torch import nn


class Sine(nn.Module):
    """ """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """

        :param input:
        :type input:
        :return:
        :rtype:
        """
        return torch.sin(
            30 * input
        )  # See SIREN paper sec. 3.2, final paragraph, and supplement Sec. 1.5 for discussion of factor 30


class Cosine(nn.Module):
    """ """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """

        :param input:
        :type input:
        :return:
        :rtype:
        """
        return torch.cos(
            30 * input
        )  # See SIREN paper sec. 3.2, final paragraph, and supplement Sec. 1.5 for discussion of factor 30


if __name__ == "__main__":

    def _main():
        from draugr.torch_utilities import to_tensor

        print(Sine()(to_tensor(1.2)))

    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 7/5/22
           """
__all__ = ["scale"]

from typing import Sequence, List

import numpy


def scale(x: Sequence, length: float) -> List[int]:
    """
    Scale points in 'x', such that distance between
    max(x) and min(x) equals to 'length'. min(x)
    will be moved to 0."""
    if isinstance(x, list) and False:
        s = float(length) / (max(x) - min(x)) if x and max(x) - min(x) != 0 else length
        min_x = min(x)
    # elif type(x) is range:
    #  s = length
    else:
        s = (
            float(length) / (numpy.max(x) - numpy.min(x))
            if len(x) and numpy.max(x) - numpy.min(x) != 0
            else length
        )
        min_x = numpy.min(x)

    return [int((i - min_x) * s) for i in x]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
NOT VERY USEFUL! as an equals chain does the same thing as this function 

           Created on 12-05-2021
           """

__all__ = ["BroadcastNone"]

import sorcery
from sorcery.core import FrameInfo


class BroadcastNone(type(None).__class__):
    """description"""

    @sorcery.spell
    def __new__(frame_info: FrameInfo, cls):
        num_assignments = len(frame_info.assigned_names(allow_one=True)[0])
        if num_assignments > 1:
            return (None,) * num_assignments
        return


if __name__ == "__main__":

    def asd213aa() -> None:
        """
        :rtype: None
        """

        assd = BroadcastNone()
        ajis, sad, asd = BroadcastNone()

        print(assd)
        print(ajis, sad, asd)

    def asd213a2a() -> None:
        """
        :rtype: None
        """

        # ajis = sad = asd = BroadcastNone() # DOES NOT WORK!
        ajis = sad = asd = None

        print(ajis, sad, asd)

    asd213aa()
    asd213a2a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/08/2020
           """

__all__ = []

import pytest

from warg import ensure_in_sys_path, find_nearest_ancestral_relative

ensure_in_sys_path(find_nearest_ancestral_relative("draugr").parent)


def test_import():
    import draugr

    print(draugr.__version__)


@pytest.mark.skip
def test_import_dlib_utilities():
    import draugr.dlib_utilities

    print(draugr.dlib_utilities.__author__)


def test_import_torch_utilities():
    import draugr.torch_utilities

    print(draugr.torch_utilities.__author__)


def test_import_numpy_utilities():
    import draugr.numpy_utilities

    print(draugr.numpy_utilities.__author__)


def test_import_multiprocessing_utilities():
    import draugr.multiprocessing_utilities

    print(draugr.multiprocessing_utilities.__author__)


def test_import_pandas_utilities():
    import draugr.pandas_utilities

    print(draugr.pandas_utilities.__author__)


def test_import_opencv_utilities():
    import draugr.opencv_utilities

    print(draugr.opencv_utilities.__author__)


def test_import_drawers():
    import draugr.drawers

    print(draugr.drawers.__author__)


def test_import_writers():
    import draugr.writers

    print(draugr.writers.__author__)


def test_import_scipy_utilities():
    import draugr.scipy_utilities

    print(draugr.scipy_utilities.__author__)


@pytest.mark.skip
def test_import_matlab_utilities():
    import draugr.matlab_utilities

    print(draugr.matlab_utilities.__author__)


@pytest.mark.skip
def test_import_tensorboard_utilities():
    import draugr.tensorboard_utilities

    print(draugr.tensorboard_utilities.__author__)


def test_import_extensions():
    import draugr.extensions

    print(draugr.extensions.__author__)


def test_import_metrics():
    import draugr.metrics

    print(draugr.metrics.__author__)


def test_import_generators():
    import draugr.python_utilities.generators

    print(draugr.python_utilities.generators.__author__)


def test_import_os_utilities():
    import draugr.os_utilities

    print(draugr.os_utilities.__author__)


def test_import_python_utilities():
    import draugr.python_utilities

    print(draugr.python_utilities.__author__)


def test_import_random_utilities():
    import draugr.random_utilities

    print(draugr.random_utilities.__author__)


def test_import_stopping():
    import draugr.stopping

    print(draugr.stopping.__author__)


def test_import_visualisation():
    import draugr.visualisation

    print(draugr.visualisation.__author__)


def test_import_entry_points():
    import draugr.entry_points

    print(draugr.entry_points.__author__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 11/30/22
           """

__all__ = ["get_video_sources"]


from pathlib import Path
from typing import Set


from warg import is_windows


def get_video_sources() -> Set[int]:
    """
    video_source_indices

    :return:
    :rtype:
    """
    if is_windows():
        from os_utilities.windows_utilities.usb_devices import get_windows_usb_devices

        usb_devices = get_windows_usb_devices()
        return set([i for i, _ in enumerate(usb_devices)])

    char_devices = [f for f in Path("/dev").iterdir() if f.is_char_device()]
    return set(
        sorted(
            [int(dev.name[-1]) for dev in char_devices if dev.name.startswith("video")]
        )
    )


if __name__ == "__main__":
    print(get_video_sources())

    def iujhasd():
        from opencv_utilities import AsyncVideoStream

        for c in get_video_sources():
            for a in AsyncVideoStream(c):
                print(a)
                break

    iujhasd()



__all__ = ["BaseFourierFeatureMLP"]
__doc__ = (
    r""""Module containing models based upon the Fourier Feature Networks template."""
)

import math
from typing import List, Optional

import torch
from torch import nn


class BaseFourierFeatureMLP(nn.Module):
    """MLP which uses Fourier features as a preprocessing step."""

    def __init__(
        self,
        num_inputs: int,
        num_outputs: int,
        a_values: Optional[torch.Tensor],
        b_values: Optional[torch.Tensor],
        layer_channels: List[int],
    ):
        """Constructor.

        Args:
            num_inputs (int): Number of dimensions in the input
            num_outputs (int): Number of dimensions in the output
            a_values (torch.Tensor): a values for encoding
            b_values (torch.Tensor): b values for encoding
            layer_channels (List[int]): Number of channels per layer.
        """
        nn.Module.__init__(self)
        self.params = {
            "num_inputs": num_inputs,
            "num_outputs": num_outputs,
            "a_values": None if a_values is None else a_values.tolist(),
            "b_values": None if b_values is None else b_values.tolist(),
            "layer_channels": layer_channels,
        }
        self.num_inputs = num_inputs
        if b_values is None:
            self.a_values = None
            self.b_values = None
            num_inputs = num_inputs
        else:
            assert b_values.shape[0] == num_inputs
            assert a_values.shape[0] == b_values.shape[1]
            self.a_values = nn.Parameter(a_values, requires_grad=False)
            self.b_values = nn.Parameter(b_values, requires_grad=False)
            num_inputs = b_values.shape[1] * 2

        self.layers = nn.ModuleList()
        for num_channels in layer_channels:
            self.layers.append(nn.Linear(num_inputs, num_channels))
            num_inputs = num_channels

        self.layers.append(nn.Linear(num_inputs, num_outputs))

        self.use_view = False
        self.keep_activations = False
        self.activations = []

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predicts outputs from the provided uv input."""
        if self.b_values is None:
            output = inputs
        else:
            # NB: the below should be 2*math.pi, but the values
            # coming in are already in the range of -1 to 1 or
            # 0 to 2, so we want to keep the range so that it does
            # not exceed 2pi
            encoded = (math.pi * inputs) @ self.b_values
            output = torch.cat(
                [self.a_values * encoded.cos(), self.a_values * encoded.sin()], dim=-1
            )

        self.activations.clear()
        for layer in self.layers[:-1]:
            output = torch.relu(layer(output))

        if self.keep_activations:
            self.activations.append(output.detach().cpu().numpy())

        output = self.layers[-1](output)
        return output

    def save(self, path: str):
        """Saves the model to the specified path.

        Args:
            path (str): Path to the model file on disk
        """
        state_dict = self.state_dict()
        state_dict["type"] = "fourier"
        state_dict["params"] = self.params
        torch.save(state_dict, path)


class MLP(BaseFourierFeatureMLP):
    """Unencoded FFN, essentially a standard MLP."""

    def __init__(
        self, num_inputs: int, num_outputs: int, num_layers=3, num_channels=256
    ):
        """Constructor.

        Args:
            num_inputs (int): Number of dimensions in the input
            num_outputs (int): Number of dimensions in the output
            num_layers (int, optional): Number of layers in the MLP.
                                        Defaults to 4.
            num_channels (int, optional): Number of channels in the MLP.
                                          Defaults to 256.
        """
        super().__init__(
            num_inputs, num_outputs, None, None, [num_channels] * num_layers
        )


class BasicFMLP(BaseFourierFeatureMLP):
    """Basic version of FFN in which inputs are projected onto the unit circle."""

    def __init__(
        self, num_inputs: int, num_outputs: int, num_layers=3, num_channels=256
    ):
        """Constructor.

        Args:
            num_inputs (int): Number of dimensions in the input
            num_outputs (int): Number of dimensions in the output
            num_layers (int, optional): Number of layers in the MLP.
                                        Defaults to 4.
            num_channels (int, optional): Number of channels in the MLP.
                                          Defaults to 256.
        """
        a_values = torch.ones(num_inputs)
        b_values = torch.eye(num_inputs)
        super().__init__(
            num_inputs,
            num_outputs,
            a_values,
            b_values,
            [num_channels] * num_layers,
        )


class PositionalFMLP(BaseFourierFeatureMLP):
    """Version of FFN with positional encoding."""

    def __init__(
        self,
        num_inputs: int,
        num_outputs: int,
        max_log_scale: float,
        num_layers=3,
        num_channels=256,
        embedding_size=256,
    ):
        """Constructor.

        Args:
            num_inputs (int): Number of dimensions in the input
            num_outputs (int): Number of dimensions in the output
            max_log_scale (float): Maximum log scale for embedding
            num_layers (int, optional): Number of layers in the MLP.
                                        Defaults to 4.
            num_channels (int, optional): Number of channels in the MLP.
                                          Defaults to 256.
            embedding_size (int, optional): The size of the feature embedding.
                                            Defaults to 256.
        """
        b_values = self._encoding(max_log_scale, embedding_size, num_inputs)
        a_values = torch.ones(b_values.shape[1])
        super().__init__(
            num_inputs,
            num_outputs,
            a_values,
            b_values,
            [num_channels] * num_layers,
        )

    @staticmethod
    def _encoding(max_log_scale: float, embedding_size: int, num_inputs: int):
        """Produces the encoding b_values matrix."""
        embedding_size = embedding_size // num_inputs
        frequencies_matrix = 2.0 ** torch.linspace(0, max_log_scale, embedding_size)
        frequencies_matrix = frequencies_matrix.reshape(-1, 1, 1)
        frequencies_matrix = torch.eye(num_inputs) * frequencies_matrix
        frequencies_matrix = frequencies_matrix.reshape(-1, num_inputs)
        frequencies_matrix = frequencies_matrix.transpose(0, 1)
        return frequencies_matrix


class GaussianFMLP(BaseFourierFeatureMLP):
    """Version of a FFN using a full Gaussian matrix for encoding."""

    def __init__(
        self,
        num_inputs: int,
        num_outputs: int,
        sigma: float,
        num_layers=3,
        num_channels=256,
        embedding_size=256,
    ):
        """Constructor.

        Args:
            num_inputs (int): Number of dimensions in the input
            num_outputs (int): Number of dimensions in the output
            sigma (float): Standard deviation of the Gaussian distribution
            num_layers (int, optional): Number of layers in the MLP.
                                        Defaults to 4.
            num_channels (int, optional): Number of channels in the MLP.
                                          Defaults to 256.
            embedding_size (int, optional): Number of frequencies to use for
                                             the encoding. Defaults to 256.
        """
        b_values = torch.normal(0, sigma, size=(num_inputs, embedding_size))
        a_values = torch.ones(b_values.shape[1])
        super().__init__(
            num_inputs,
            num_outputs,
            a_values,
            b_values,
            [num_channels] * num_layers,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 7/13/22
           """

__all__ = []

from pathlib import Path
from typing import Sequence, Mapping, Union, MutableMapping

import numpy
from PIL import Image
from draugr.writers import (
    Writer,
    ImageWriterMixin,
    HistogramWriterMixin,
    BarWriterMixin,
    LineWriterMixin,
    SpectrogramWriterMixin,
    FigureWriterMixin,
)
from matplotlib import pyplot
from matplotlib.figure import Figure

from warg import passes_kws_to


class MatPlotLibWriter(
    Writer,
    ImageWriterMixin,
    HistogramWriterMixin,
    BarWriterMixin,
    LineWriterMixin,
    SpectrogramWriterMixin,
    FigureWriterMixin,
):
    """
    description
    """

    @passes_kws_to(pyplot.savefig)
    def __init__(self, path: Path, format_: str = "png", **kwargs: MutableMapping):
        super().__init__()
        self._path = path
        self._format = format_
        self.kws = kwargs
        self._open()

    def _close(self, exc_type=None, exc_val=None, exc_tb=None) -> None:
        del self.scalars

    def _open(self):
        self.scalars = {}
        return self

    def _scalar(self, tag: str, value: float, step: int) -> None:
        if tag not in self.scalars:
            self.scalars[tag] = []
        self.scalars[tag] += value
        fig: Figure = pyplot.imshow(value)
        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()

    @passes_kws_to(pyplot.imshow)
    def image(
        self,
        tag: str,
        data: Union[numpy.ndarray, Image.Image],
        step: int,
        *,
        dataformats: str = "NCHW",
        **kwargs: MutableMapping,
    ) -> None:
        """
        Plot an image in matplotlib.

        :param tag:
        :type tag:
        :param data:
        :type data:
        :param step:
        :type step:
        :param dataformats:
        :type dataformats:
        :param kwargs:
        :type kwargs:
        """
        fig = pyplot.imshow(data, **kwargs)
        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()

    @passes_kws_to(pyplot.hist)
    def histogram(
        self, tag: str, values: list, step: int, **kwargs: MutableMapping
    ) -> None:
        """
        Plot a histogram in matplotlib.

        :param tag:
        :type tag:
        :param values:
        :type values:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:
        """
        fig = pyplot.hist(values, **kwargs)

        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()

    def bar(
        self,
        tag: str,
        values: list,
        step: int,
        y_error=None,
        x_labels=None,
        y_label="Probability",
        x_label="Action Categorical Distribution",
        **kwargs: MutableMapping,
    ) -> None:
        """
        Plot a bar chart in matplotlib.

        :param tag:
        :type tag:
        :param values:
        :type values:
        :param step:
        :type step:
        :param y_error:
        :type y_error:
        :param x_labels:
        :type x_labels:
        :param y_label:
        :type y_label:
        :param x_label:
        :type x_label:
        :param kwargs:
        :type kwargs:
        """
        fig = pyplot.bar(values)
        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()

    def line(
        self,
        tag: str,
        values: list,
        step: int,
        x_labels: Sequence = None,
        y_label: str = "Magnitude",
        x_label: str = "Sequence",
        plot_kws: Mapping = None,
        **kwargs: MutableMapping,
    ) -> None:
        """
        Plot a line chart in matplotlib.

        :param tag:
        :type tag:
        :param values:
        :type values:
        :param step:
        :type step:
        :param x_labels:
        :type x_labels:
        :param y_label:
        :type y_label:
        :param x_label:
        :type x_label:
        :param plot_kws:
        :type plot_kws:
        :param kwargs:
        :type kwargs:
        """
        fig = pyplot.plot(values)
        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()

    def spectrogram(
        self,
        tag: str,
        values: list,
        sample_rate: int,
        step: int,
        num_fft: int = 512,
        x_labels: Sequence = None,
        y_label: str = "Magnitude",
        x_label: str = "Sequence",
        plot_kws: Mapping = None,
        **kwargs: MutableMapping,
    ) -> None:
        """
        Plot a spectrogram in matplotlib.

        :param tag:
        :type tag:
        :param values:
        :type values:
        :param sample_rate:
        :type sample_rate:
        :param step:
        :type step:
        :param num_fft:
        :type num_fft:
        :param x_labels:
        :type x_labels:
        :param y_label:
        :type y_label:
        :param x_label:
        :type x_label:
        :param plot_kws:
        :type plot_kws:
        :param kwargs:
        :type kwargs:
        """
        fig = pyplot.specgram(values, Fs=sample_rate, NFFT=num_fft)
        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()

    def figure(
        self, tag: str, figure: Figure, step: int, **kwargs: MutableMapping
    ) -> None:
        """
        Plot a figure in matplotlib.

        :param tag:
        :type tag:
        :param figure:
        :type figure:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:
        """
        fig = pyplot.figure(figure)
        fig.savefig(
            (self._path / f"{tag}_{step}").with_suffix(self._format),
            format=self._format,
            **self.kws,
        )
        fig.close()


if __name__ == "__main__":
    with MatPlotLibWriter() as writer:
        writer.scalar("test", 1.0, 0)
        writer.scalar("test", 2.0, 1)
        writer.scalar("test", 3.0, 2)

        writer.image("test", numpy.random.rand(3, 3, 3), 0)
        writer.image("test", numpy.random.rand(3, 3, 3), 1)

        writer.histogram("test", [1, 2, 3, 4, 5], 0)
        writer.histogram("test", [1, 2, 3, 4, 5], 1)

        writer.bar("test", [1, 2, 3, 4, 5], 0)
        writer.bar("test", [1, 2, 3, 4, 5], 1)

        writer.line("test", [1, 2, 3, 4, 5], 0)
        writer.line("test", [1, 2, 3, 4, 5], 1)

        writer.spectrogram("test", [1, 2, 3, 4, 5], 1, 0)
        writer.spectrogram("test", [1, 2, 3, 4, 5], 1, 1)

        writer.figure("test", pyplot.figure(), 0)
        writer.figure("test", pyplot.figure(), 1)



from typing import Callable

import cv2
from draugr.opencv_utilities import ButtonTypeEnum

from warg import sink

__all__ = ["add_checkbox"]


def add_checkbox(
    checkbox_name: str,
    *,
    initial_button_state: bool = False,
    callback: Callable = sink,
    new_line: bool = True
) -> None:
    """

    :param checkbox_name:
    :type checkbox_name:
    :param initial_button_state:
    :type initial_button_state:
    :param new_line:
    :type new_line:
    :param callback: Callback function to be called when the checkbox is changed.
    :return: None
    """
    checked = 0
    if initial_button_state:
        checked = 1

    type_ = ButtonTypeEnum.check_box.value

    if new_line:
        type_ |= ButtonTypeEnum.button_bar.value

    cv2.createButton(
        checkbox_name,
        # window_name,
        callback,
        None,
        type_,
        checked,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 8/24/22
           """

__all__ = []

from typing import MutableMapping, Sequence

import torch
from draugr.torch_utilities.architectures.mlp import MLP
from torch import nn
from torch.nn import functional


class RecurrentCategoricalMLP(MLP):
    """
    Recurrent model base the MLP base model
    """

    def __init__(self, r_hidden_layers=10, **kwargs: MutableMapping):
        super().__init__(**kwargs)
        self._r_hidden_layers = r_hidden_layers
        self._r_input_shape = self._output_shape + r_hidden_layers

        self.hidden = nn.Linear(
            self._r_input_shape, r_hidden_layers, bias=self._use_bias
        )
        self.out = nn.Linear(self._r_input_shape, r_hidden_layers, bias=self._use_bias)

        self._prev_hidden_x = torch.zeros(r_hidden_layers)

    def forward(self, x: Sequence, **kwargs: MutableMapping):
        """

        :param x:
        :type x:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        x = super().forward(x, **kwargs)
        combined = torch.cat((x, self._prev_hidden_x), 1)
        out_x = self.out(combined)
        hidden_x = self.hidden(combined)
        self._prev_hidden_x = hidden_x

        return functional.log_softmax(out_x, dim=-1)


class ExposedRecurrentCategoricalMLP(RecurrentCategoricalMLP):
    """
    Exposed Variant of Recurrent model base the MLP base model

    """

    def forward(self, x: Sequence, hidden_x: torch.Tensor, **kwargs: MutableMapping):
        """

        :param x:
        :type x:
        :param hidden_x:
        :type hidden_x:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        self._prev_hidden_x = hidden_x
        out_x = super().forward(x, self._prev_hidden_x, **kwargs)

        return functional.log_softmax(out_x, dim=-1), self._prev_hidden_x



import win32api
import win32con
import win32gui


def press_key(hwnd, s, key, start_sec, hold_sec):
    """
    send a keyboard input to the given window

    :param hwnd:
    :type hwnd:
    :param s:
    :type s:
    :param key:
    :type key:
    :param start_sec:
    :type start_sec:
    :param hold_sec:
    :type hold_sec:
    :return:
    :rtype:
    """
    priority = 2
    duration = start_sec + hold_sec

    s.enter(
        start_sec,
        priority,
        win32api.SendMessage,
        argument=(hwnd, win32con.WM_KEYDOWN, key, 0),
    )
    s.enter(
        duration,
        priority,
        win32api.SendMessage,
        argument=(hwnd, win32con.WM_KEYUP, key, 0),
    )

    # win32gui.SetForegroundWindow(hwnd)
    # win32api.SendMessage(hwnd, win32con.WM_KEYDOWN, key, 0)
    # sleep(sec)
    # win32api.SendMessage(hwnd, win32con.WM_KEYUP, key, 0)


def press_key_foreground(hwnd, s, key, start_sec, hold_sec):
    """
    # send a keyboard input to the given window
    """
    priority = 2
    foreground_time = 0.15
    duration = start_sec + hold_sec

    s.enter(
        start_sec - foreground_time,
        priority,
        win32gui.SetForegroundWindow,
        argument=(hwnd,),
    )
    s.enter(
        start_sec,
        priority,
        win32api.SendMessage,
        argument=(hwnd, win32con.WM_KEYDOWN, key, 0),
    )
    s.enter(
        duration - foreground_time,
        priority,
        win32gui.SetForegroundWindow,
        argument=(hwnd,),
    )
    s.enter(
        duration,
        priority,
        win32api.SendMessage,
        argument=(hwnd, win32con.WM_KEYUP, key, 0),
    )

    # win32gui.SetForegroundWindow(hwnd)
    # win32api.SendMessage(hwnd, win32con.WM_KEYDOWN, key, 0)
    # sleep(sec)
    # win32api.SendMessage(hwnd, win32con.WM_KEYUP, key, 0)



from sorcery import assigned_names


def upper_case_assigned_names():
    """

    :return:
    :rtype:
    """
    return {name.upper(): value for name, value in assigned_names().items()}


def lower_case_assigned_names():
    """

    :return:
    :rtype: object
    """
    return {name.lower(): value for name, value in assigned_names().items()}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = ""

from typing import Tuple

import torch
from torch import nn


class RecurrentBase(nn.Module):
    """
    A base class for Recurrent models
    """

    def __init__(self, recurrent: bool, recurrent_input_size: int, hidden_size: int):
        super().__init__()

        self._hidden_size = hidden_size
        self._recurrent = recurrent

        if recurrent:
            self.gru = nn.GRUCell(recurrent_input_size, hidden_size)
            nn.init.orthogonal_(self.gru.weight_ih.data)
            nn.init.orthogonal_(self.gru.weight_hh.data)
            self.gru.bias_ih.data.fill_(0)
            self.gru.bias_hh.data.fill_(0)

    def _forward_gru(
        self: torch.Tensor, x, hxs: torch.Tensor, masks: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if x.size(0) == hxs.size(0):
            x = hxs = self.gru(x, hxs * masks)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.reshape(T, N, x.size(1))

            # Same deal with masks
            masks = masks.reshape(T, N, 1)

            outputs = []
            for i in range(T):
                hx = hxs = self.gru(x[i], hxs * masks[i])
                outputs.append(hx)

            # assert len(outputs) == T
            # x is a (T, N, -1) tensor
            x = torch.stack(outputs, dim=0)
            # flatten
            x = x.reshape(T * N, -1)

        return x, hxs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

            TODO: Finish this off, only a placeholder for now

           Created on 17-03-2021
           """

from typing import Sequence, MutableMapping, Any

from draugr.tensorboard_utilities.exporting.event_export import TensorboardEventExporter


class TensorboardEventExporterFile(TensorboardEventExporter):
    def __init__(self, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
        super().__init__(*args, **kwargs)
        self.file = kwargs.get("file")
        if self.file is not None:
            self.file.write("")
        else:
            raise ValueError("No file provided")

    def export(self):
        self.file.write("")
        super().export()

    def close(self):
        self.file.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

TODO: UNIFY PROGRESS BARs interfaces, base on system and python context

           Created on 8/24/22
           """

__all__ = ["ETABar"]

from typing import Iterator

from draugr.python_utilities import in_ipynb
from progress.bar import Bar  # IT is shit
from progress.colors import color  # shit

from warg import passes_kws_to, drop_unused_kws


class ETABar(Bar, Iterator):
    """Progress bar that displays the estimated time of completion.
    TODO: REMOVE THIS PIECE OF SHIT!

    """

    def __next__(self):
        return self.next()

    default_suffix = "%(percent).1f%% - %(eta)ds"
    suffix = default_suffix
    message = "%(eta)"
    bar_prefix = " "
    bar_suffix = " "
    empty_fill = "∙"
    fill = "█"

    def update(self, n: int) -> None:
        """

        :param n:
        :type n:
        :return:
        :rtype:
        """
        rv = self.next(n)
        super().update()
        return rv

    @drop_unused_kws
    @passes_kws_to(Bar.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def writeln(self, line: str):
        """Writes the line to the console.

        Description:
            This method is Jupyter notebook aware, and will do the
            right thing when in that environment as opposed to being
            run from the command line.

        Args:
            line (str): The message to write
        """
        if in_ipynb():
            from IPython.display import clear_output

            clear_output(wait=True)
            self.fill = "#"
            print(line)
        else:
            Bar.writeln(self, line)

    def info(self, text: str):
        """Appends the given information to the progress bar message.

        Args:
            text (str): A status message for the progress bar.
        """
        self.suffix = f"{ETABar.default_suffix} {text}"

    def update(self):
        """ """
        filled_length = int(self.width * self.progress)
        empty_length = self.width - filled_length

        message = self.message % self
        bar = color(self.fill * filled_length, fg=self.color)
        empty = self.empty_fill * empty_length
        suffix = self.suffix % self
        line = "".join([message, self.bar_prefix, bar, empty, self.bar_suffix, suffix])
        self.writeln(line)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

__doc__ = r"""
"""

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()
# #del Path # find a way to sanitise namespace

from .accumulation import *
from .meters import *
from .metric_aggregator import *
from .metric_collection import *
from .metric_summary import *

""" Sanitize namespace # https://stackoverflow.com/questions/61278110/excluding-modules-when-importing-everything-in-init-py
import types
__all__ = [name for name, thing in globals().items()
          if not (name.startswith('_') or isinstance(thing, types.ModuleType))]
del types
"""



import subprocess
from pathlib import Path
from typing import Optional

from warg import Number
from warg import ensure_existence

FORMAT_LIST = [".mp4", ".avi", ".mkv", ".flv", ".mov"]
AUDIO_FORMAT = ".aac"

__all__ = ["extract_frames"]


def extract_frames(
    file_path: Path,
    frame_out_dir: Optional[Path] = None,
    audio_out_dir: Optional[Path] = None,
    rate: Number = 25,
    frame_format: str = "jpg",
    extract_sound: bool = True,
    ffmpeg_path: Path = "ffmpeg",
) -> None:
    """

    :param file_path:
    :type file_path:
    :param frame_out_dir:
    :type frame_out_dir:
    :param audio_out_dir:
    :type audio_out_dir:
    :param rate:
    :type rate:
    :param frame_format:
    :type frame_format:
    :param extract_sound:
    :type extract_sound:
    :param ffmpeg_path:
    :type ffmpeg_path:
    """
    root_dir = file_path.parent
    if frame_out_dir is None:
        frame_out_dir = ensure_existence(root_dir / file_path.stem / "frames")
    if audio_out_dir is None:
        audio_out_dir = ensure_existence(frame_out_dir.parent / "audio")

    if file_path.is_file() and file_path.suffix in FORMAT_LIST:
        print(f"start extracting {file_path} frames")
        subprocess.call(
            [
                str(ffmpeg_path),
                "-i",
                file_path,
                "-r",
                str(rate),
                "-f",
                "image2",
                "-y",
                "-qscale:v",
                "2",
                str(frame_out_dir / f"%d.{frame_format}"),
            ]
        )
        print(f"end extracting {file_path} frames")

        if extract_sound:
            print(f"start extracting {file_path} audio")
            subprocess.call(
                [
                    str(ffmpeg_path),
                    "-i",
                    file_path,
                    "-vn",
                    "-acodec",
                    "copy",
                    "-y",
                    str(audio_out_dir / f"track{AUDIO_FORMAT}"),
                ]
            )
            print(f"end extracting {file_path} audio")
    else:
        print(f"{file_path} is not a video file")


if __name__ == "__main__":
    # a = Path.home() / "DataWin" / "DeepFake" / "Frontier" / "Originals" / "thomas_old_high_res.mp4"

    a = (
        Path.home()
        / "SynologyDrive"
        / "Frontier"
        / "Fra Frontier"
        / "Personer"
        / "Peter AG"
        / "Peter AG 1983+1991.mp4"
    )
    ffmpeg_path = "ffmpeg"
    if False:
        ffmpeg_path = (
            Path.home()
            / "OneDrive - Alexandra Instituttet"
            / "Applications"
            / "ffmpeg-5.0-essentials_build"
            / "bin"
            / "ffmpeg.exe"
        )
    extract_frames(
        a,
        ffmpeg_path=ffmpeg_path,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 7/5/22
           """

from .terminal_image_renderer import *
from .terminal_plot import *
from .terminal_plotter import *
from .terminal_utilities import *



from enum import Enum

from sorcery import assigned_names

__all__ = [
    "PilModesEnum",
]


class PilModesEnum(Enum):
    """
      PIL pixel formats:

    RGB 24bits per pixel, 8-bit-per-channel RGB), 3 channels
    RGBA (8-bit-per-channel RGBA), 4 channels
    RGBa (8-bit-per-channel RGBA, remultiplied alpha), 4 channels
    1 - 1bpp, often for masks, 1 channel
    L - 8bpp, grayscale, 1 channel
    P - 8bpp, paletted, 1 channel
    I - 32-bit integers, grayscale, 1 channel
    F - 32-bit floats, grayscale, 1 channel
    CMYK - 8 bits per channel, 4 channels
    YCbCr - 8 bits per channel, 3 channels
    """

    OneBpp = "1"
    CMYK, F, HSV, I, L, LAB, P, RGB, RGBA, RGBX, YCbCr = assigned_names()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/20/22
           """

__all__ = ["named_parameters_numpy"]

from typing import Dict

import torch
from torch.nn import Module


def named_parameters_numpy(model: Module) -> Dict[str, torch.nn.Parameter]:
    """

    :param model:
    :type model:
    :return:
    :rtype:"""
    params_to_export = {}
    for name, param in model.named_parameters():
        params_to_export[name] = param.detach().cpu().numpy()
    return params_to_export



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

import numpy
import torch

from torch.nn import (
    Conv2d,
    Linear,
    Module,
    Sigmoid,
    Tanh,
    ReLU,
    LeakyReLU,
    Softmax,
    Softplus,
    SiLU,
    ELU,
    Identity,
    Conv1d,
    Conv3d,
)
from torch.nn.init import calculate_gain, constant_, uniform_, xavier_uniform_

__all__ = ["fan_in_init", "xavier_init", "constant_init", "normal_init", "ortho_init"]


def fan_in_init(model: Module) -> None:
    """

    :param model:
    :type model:"""
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            fan_in = m.weight.size(1)
            v = 1.0 / numpy.sqrt(fan_in)
            uniform_(m.weight, -v, v)
            constant_(m.bias, 0)


NonLinearityEnum = {
    Sigmoid: "sigmoid",
    torch.sigmoid: "sigmoid",
    Tanh: "tanh",
    torch.tanh: "tanh",
    ReLU: "relu",
    torch.relu: "relu",
    LeakyReLU: "leaky_relu",
    ELU: "elu",
    SiLU: "silu",
    Softplus: "softplus",
    Softmax: "softmax",
    torch.softmax: "softmax",
    Linear: "linear",
    Identity: "identity",
    torch.conv1d: "conv1d",
    Conv1d: "conv1d",
    torch.conv2d: "conv2d",
    Conv2d: "conv2d",
    torch.conv3d: "conv3d",
    Conv3d: "conv3d",
    torch.conv_transpose1d: "conv_transpose1d",
    torch.nn.ConvTranspose1d: "conv_transpose1d",
    torch.conv_transpose2d: "conv_transpose2d",
    torch.nn.ConvTranspose2d: "conv_transpose2d",
    torch.conv_transpose3d: "conv_transpose3d",
    torch.nn.ConvTranspose3d: "conv_transpose3d",
}


def xavier_init(model: Module, activation: NonLinearityEnum = "relu") -> None:
    """

    :param model:
    :type model:
    :param activation:
    :type activation:"""

    if not isinstance(activation, str):
        activation = NonLinearityEnum[activation].value

    gain = calculate_gain(activation)
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            xavier_uniform_(m.weight, gain=gain)
            constant_(m.bias, 0)


def constant_init(model: Module, constant: float = 1) -> None:
    """

    :param model:
    :type model:
    :param constant:
    :type constant:"""
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            constant_(m.weight, constant)
            constant_(m.bias, constant)


def normal_init(model: Module, mean: float = 0, std: float = 1.0) -> None:
    """

    :param model:
    :param mean:
    :param std:
    :return:
    """
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            torch.nn.init.normal_(m.weight, mean, std)
            torch.nn.init.normal_(m.bias, mean, std)


def ortho_init(module: Module, gain: float = 1) -> None:
    """
    Initialize the weights of a module with orthogonal initialization.
    """
    for m in module.modules():
        if isinstance(m, (Linear, Conv2d)):
            torch.nn.init.orthogonal_(m.weight, gain=gain)
            if m.bias is not None:
                m.bias.data.fill_(0.0)


if __name__ == "__main__":
    a = Linear(3, 3)
    fan_in_init(a)
    xavier_init(a)
    constant_init(a)
    ortho_init(a)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

            TODO: Finish this off, only a placeholder for now

           Created on 17-03-2021
           """

from typing import Sequence, MutableMapping, Any

from draugr.tensorboard_utilities.exporting.event_export import TensorboardEventExporter


class TensorboardEventExporterDatabase(TensorboardEventExporter):
    def __init__(self, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
        super().__init__(*args, **kwargs)
        postgres_db = kwargs.get("postgres_db")
        if postgres_db is not None:
            self.database = postgres_db
        else:
            raise ValueError("No postgres_db provided")

    def export(self, database):
        self.database = database
        super().export(database)

    def export_event(self, event):
        self.database.add_event(event)
        super().export_event(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 27/06/2020
           """

import torch
from torch.nn import functional

__all__ = ["get_num_samples", "get_archetypes", "archetypal_loss"]


def get_num_samples(
    targets: torch.Tensor, num_classes: torch.Tensor, dtype: torch.dtype = None
) -> torch.Tensor:
    """

    :param targets:
    :type targets:
    :param num_classes:
    :type num_classes:
    :param dtype:
    :type dtype:
    :return:
    :rtype:
    """
    batch_size = targets.size(0)
    with torch.no_grad():
        ones = torch.ones_like(targets, dtype=dtype)
        num_samples = ones.new_zeros((batch_size, num_classes))
        num_samples.scatter_add_(1, targets.type(torch.int64), ones)
    return num_samples


def get_archetypes(
    embeddings: torch.FloatTensor, targets: torch.LongTensor, num_classes: int
) -> torch.FloatTensor:
    """Compute the archetypes (the mean vector of the embedded training/support
    points belonging to its class) for each classes in the task.

    Parameters
    ----------
    embeddings : `torch.FloatTensor` instance
        A tensor containing the embeddings of the support points. This tensor
        has shape `(batch_size, num_examples, embedding_size)`.

    targets : `torch.LongTensor` instance
        A tensor containing the targets of the support points. This tensor has
        shape `(batch_size, num_examples)`.

    num_classes : int
        Number of classes in the task.

    Returns
    -------
    archetypes : `torch.FloatTensor` instance
        A tensor containing the archetypes for each class. This tensor has shape
        `(batch_size, num_classes, embedding_size)`.
    """
    batch_size, embedding_size = embeddings.size(0), embeddings.size(-1)

    num_samples = get_num_samples(targets, num_classes, dtype=embeddings.dtype)
    num_samples.unsqueeze_(-1)
    num_samples = torch.max(num_samples, torch.ones_like(num_samples))

    prototypes = embeddings.new_zeros((batch_size, num_classes, embedding_size))
    indices = targets.unsqueeze(-1).expand_as(embeddings).type(torch.int64)
    prototypes.scatter_add_(1, indices, embeddings).div_(num_samples)

    return prototypes


def archetypal_loss(
    archetypes: torch.FloatTensor,
    embeddings: torch.FloatTensor,
    targets: torch.LongTensor,
    **kwargs
) -> torch.FloatTensor:
    """Compute the loss (i.e. negative log-likelihood) for the archetypal
    network, on the test/query points.

    Parameters
    ----------
    archetypes : `torch.FloatTensor` instance
        A tensor containing the archetypes for each class. This tensor has shape
        `(batch_size, num_classes, embedding_size)`.

    embeddings : `torch.FloatTensor` instance
        A tensor containing the embeddings of the query points. This tensor has
        shape `(batch_size, num_examples, embedding_size)`.

    targets : `torch.LongTensor` instance
        A tensor containing the targets of the query points. This tensor has
        shape `(batch_size, num_examples)`.

    Returns
    -------
    loss : `torch.FloatTensor` instance
        The negative log-likelihood on the query points.
    """
    squared_distances = torch.sum(
        (archetypes.unsqueeze(2) - embeddings.unsqueeze(1)) ** 2, dim=-1
    )
    return functional.cross_entropy(-squared_distances, targets, **kwargs)



__all__ = ["positive_int_hash"]

import zlib

import numpy

MAX_INT_32 = numpy.iinfo(numpy.int32).max


def positive_int_hash(s: str, limit_size=MAX_INT_32) -> int:
    pos_int = zlib.adler32(s.encode("utf-8")) & 0xFFFFFFFF
    if limit_size:  # Wrap around
        pos_int %= limit_size
    return pos_int



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-01-2021
           """

__all__ = ["Drawer"]

from abc import abstractmethod
from typing import Sequence, MutableMapping, Any

from warg import drop_unused_kws


class Drawer(object):
    """
    Abstract class for drawing representations of data
    """

    @drop_unused_kws
    def __init__(self, verbose: bool = False):
        self._verbose = verbose

    @abstractmethod
    def draw(self, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]) -> None:
        """description"""
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/01/2020
           """

# with open(Path(__file__).parent / "README.md", "r") as this_init_file:
#    __doc__ += this_init_file.read()  # .replace("#", "")  # .encode("ascii", "ignore")


from .counter_filter import *
from .display import *
from .function_wrappers import *
from .generators import *
from .http import *
from .iterators import *
from .matrix import *
from .platform_context import *
from .sockets import *
from .torch_like_channel_transformation import *
from .styling import *



import subprocess
from pathlib import Path
from typing import Optional

from draugr.ffmpeg_utilities.extract import AUDIO_FORMAT

from warg import Number, identity
from warg import ensure_existence

__all__ = ["merge_frames"]


def get_frame_format(frames_dir: Path, formats=(".jpg", ".png")) -> str:
    """

    :param frames_dir:
    :type frames_dir:
    :return:
    :rtype:
    """
    for file_ in frames_dir.iterdir():
        if file_.is_file():
            suffix = file_.suffix.lower()
            if suffix in formats:
                return suffix


def merge_frames(
    frames_dir: Path,
    merge_audio: bool = True,
    audio_dir: Optional[Path] = None,
    merge_dir: Optional[Path] = None,
    merge_rate: Number = 25,
    ffmpeg_path: Path = "ffmpeg",
) -> None:
    """merges frame in to video

    :param frames_dir:
    :type frames_dir:
    :param merge_audio:
    :type merge_audio:
    :param audio_dir:
    :type audio_dir:
    :param merge_dir:
    :type merge_dir:
    :param merge_rate:
    :type merge_rate:
    :param ffmpeg_path:
    :type ffmpeg_path:
    """
    postfix = ""  # "_00"
    vid_dir = frames_dir.parent

    if audio_dir is None:
        audio_dir = vid_dir / "audio"

    if merge_dir is None:
        merge_dir = ensure_existence(vid_dir / "merge", sanitisation_func=identity)

    shortname = vid_dir.name

    # a = f"{shortname}-%05d{postfix}{get_frame_format(frames_dir)}"  # 05d is for 5 digits in ids
    # a = "brandt.mp4-%*_00.png"
    a = f"%05d{get_frame_format(frames_dir)}"
    temp_out = str(merge_dir / "temp.mp4")
    subprocess.call(
        [
            str(ffmpeg_path),
            "-r",
            str(merge_rate),
            "-pattern_type",
            "glob",
            "-i",
            str(frames_dir / f"*{get_frame_format(frames_dir)}"),
            "-y",
            "-c:v",
            "libx264",
            "-vf",
            "fps=25,format=yuv420p",
            temp_out,
        ]
    )

    a = []

    if merge_audio:
        sound_dir = (audio_dir / "track").with_suffix(f".{AUDIO_FORMAT.lstrip('.')}")
        if sound_dir.exists():
            a.extend(["-i", str(sound_dir)])
        else:
            print(f"No audio found in {sound_dir}")

    subprocess.call(
        [
            str(ffmpeg_path),
            "-i",
            temp_out,
            *a,
            "-vcodec",
            "copy",
            "-acodec",
            "copy",
            "-y",
            str((merge_dir / "out").with_suffix(f".{'mp4'.lstrip('.')}")),
        ]
    )


if __name__ == "__main__":
    merge_frames(
        Path.home() / "DataWin" / "frames",
        ffmpeg_path=Path.home()
        / "OneDrive - Alexandra Instituttet"
        / "Applications"
        / "ffmpeg-5.0-essentials_build"
        / "bin"
        / "ffmpeg.exe",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""
           Render images on the command line

           Created on 7/5/22
           """

import argparse
from pathlib import Path
from typing import Any, List, Tuple

import numpy
from PIL import Image

from warg import color, Number

__all__ = ["render_file", "terminalise_image", "terminal_render_image"]


def get_pixel(col: Tuple) -> Any:
    """
    Get the pixel value for a given colour.

    :param col:"""
    if isinstance(col, (float, int)) or len(col) == 1:
        col = (col, col, col)
    return color("  ", bg=f"rgb({int(col[0])}, {int(col[1])}, {int(col[2])})")


def terminal_render_image(
    pixels: numpy.ndarray, scale: Tuple, max_val: Number = None
) -> List[List]:
    """
    Render an image on the command line."""
    # first of all scale the image to the scale 'tuple'
    if len(pixels.shape) < 3:
        pixels = numpy.array([[pixel] for pixel in pixels])
        num_channels = 1
    else:
        num_channels = pixels.shape[2]

    if max_val is None:
        max_val = numpy.max(pixels)
        if max_val <= 1:
            pixels = pixels * 255
        else:
            pixels = pixels * (255 / max_val)
    else:
        pixels = pixels * (255 / max_val)

    image_size = pixels.shape[:2]
    block_size = (image_size[0] / scale[0], image_size[1] / scale[1])

    blocks = []
    y = 0
    while y < image_size[0]:
        x = 0
        block_col = []
        while x < image_size[1]:
            # get a block, reshape in into an Nx3 matrix and then get average of each column
            block_col.append(
                pixels[int(y) : int(y + block_size[0]), int(x) : int(x + block_size[1])]
                .reshape(-1, num_channels)
                .mean(axis=0)
            )
            x += block_size[1]
        blocks.append(block_col)
        y += block_size[0]
    output = [[get_pixel(block) for block in row] for row in blocks]
    return output


def terminalise_image(output):
    """

    joins nested str lists with newlines

    :return:"""
    return "\n".join(["".join(row) for row in output])


def get_image(path: Path):
    """
    Get an image from a path."""
    img = numpy.asarray(Image.open(path))
    if img.shape[2] > 3:
        return numpy.array([[pixel[:3] for pixel in row] for row in img])
    return img


def render_file(path: Path, scale=(60, 60)):
    """
    Render an image file on the command line."""
    image = get_image(path)
    output = terminal_render_image(image, scale)
    print(terminalise_image(output))


def entry_point() -> None:
    """
    Entry point for the command line interface.

    :rtype: None
    """
    parser = argparse.ArgumentParser(description="Render images on the command line")
    parser.add_argument("path", metavar="path", type=str, help="the image path")
    parser.add_argument(
        "--width",
        dest="width",
        default=60,
        type=int,
        help="width of the rendered image (default 60 pixels)",
    )
    parser.add_argument(
        "--height",
        dest="height",
        default=60,
        type=int,
        help="height of the rendered image (default 60 pixels)",
    )
    args = parser.parse_args()
    render_file(args.path, (args.height, args.width))


if __name__ == "__main__":
    output = terminal_render_image(numpy.random.random((60, 60, 3)) * 5, (60, 60))
    print(terminalise_image(output))
    # render_file(Path.home() / "OneDrive" / "Billeder" / "pompey.jpg", scale=(20, 20))



from typing import Callable

import cv2
from draugr.opencv_utilities import ButtonTypeEnum

from warg import sink

__all__ = ["add_button"]


def add_button(
    checkbox_name: str, *, callback: Callable = sink, new_line: bool = True
) -> None:
    """
    :param checkbox_name:
    :type checkbox_name:
    :param new_line:
    :type new_line:
    :param callback: Callback function to be called when the button is pressed.
    :return: None
    """

    type_ = ButtonTypeEnum.push_button.value

    if new_line:
        type_ |= ButtonTypeEnum.button_bar.value

    cv2.createButton(checkbox_name, callback, None, type_, 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-03-2021
           """

from draugr.writers import Writer


class SqliteWriter(Writer):
    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        pass

    def _open(self):
        pass

    def _scalar(self, tag: str, value: float, step: int):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 10/3/22
           """

__all__ = ["RBFLayer"]

import numpy
import torch
from torch import nn


class RBFLayer(nn.Module):
    """Transforms incoming data using a given radial basis function.
    - Input: (1, N, in_features) where N is an arbitrary batch size
    - Output: (1, N, out_features) where N is an arbitrary batch size"""

    def __init__(self, in_features, out_features):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.centres = nn.Parameter(torch.Tensor(out_features, in_features))
        self.sigmas = nn.Parameter(torch.Tensor(out_features))
        self.reset_parameters()

        self.freq = nn.Parameter(numpy.pi * torch.ones((1, self.out_features)))

    def reset_parameters(self):
        """ """
        nn.init.uniform_(self.centres, -1, 1)
        nn.init.constant_(self.sigmas, 10)

    def forward(self, input):
        """

        :param input:
        :type input:
        :return:
        :rtype:
        """
        input = input[0, ...]
        size = (input.size(0), self.out_features, self.in_features)
        x = input.unsqueeze(1).expand(size)
        c = self.centres.unsqueeze(0).expand(size)
        distances = (x - c).pow(2).sum(-1) * self.sigmas.unsqueeze(0)
        return self.gaussian(distances).unsqueeze(0)

    def gaussian(self, alpha):
        """

        :param alpha:
        :type alpha:
        :return:
        :rtype:
        """
        phi = torch.exp(-1 * alpha.pow(2))
        return phi


if __name__ == "__main__":

    def _main():
        ...

    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/2/22
           """

__all__ = ["linspace"]

import torch


def linspace(start: torch.Tensor, stop: torch.Tensor, num_samples: int) -> torch.Tensor:
    """Generalization of linspace to arbitrary tensors.

    Args:
        start (torch.Tensor): Minimum 1D tensor. Same length as stop.
        stop (torch.Tensor): Maximum 1D tensor. Same length as start.
        num_samples (int): The number of samples to take from the linear range.

    Returns:
        torch.Tensor: (D, num_samples) tensor of linearly interpolated
                      samples.
    """
    return start.unsqueeze(-1) + (
        torch.linspace(0, 1, num_samples).unsqueeze(0)  # samples
        * (stop - start).unsqueeze(-1)  # diff
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

from .architecture import *
from .distributional import *
from .experimental import *
from .mlp import *
from .mlp_variants import *
from .recurrent import *
from .vae import *



from typing import List, Dict


def get_windows_usb_devices(
    root_usb_key: str = r"SYSTEM\CurrentControlSet\Enum\USB",
    usb_service_type: str = "usbvideo",  # TODO: ENUMIFY service_type
) -> List[Dict[str, str]]:
    try:
        import winreg
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError((e, "missing winreg"))

    WIN_ERROR_CODE_NO_DATA = 259

    usb_devices = []

    sub_keys = []
    with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, root_usb_key) as root_usb:
        index = 0
        while True:
            try:
                subkey = winreg.EnumKey(root_usb, index)
                if subkey.startswith("VID"):
                    sub_keys.append(root_usb_key + "\\" + str(subkey))
                index += 1
            except OSError as e:
                if e.winerror == WIN_ERROR_CODE_NO_DATA:  # No more data is available
                    break
                elif e.winerror == 234:  # more data is available
                    index += 1
                    continue
                raise e

    sub_sub_keys = []
    for sub_key in sub_keys:
        with winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            sub_key,
        ) as vid_key:
            index = 0
            while True:
                try:
                    sub_sub_keys.append(sub_key + "\\" + winreg.EnumKey(vid_key, index))
                    index += 1
                except OSError as e:
                    if (
                        e.winerror == WIN_ERROR_CODE_NO_DATA
                    ):  # No more data is available
                        break
                    elif e.winerror == 234:  # more data is available
                        index += 1
                        continue
                    raise e

    for sub_sub_key in sub_sub_keys:
        with winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            sub_sub_key,
        ) as dev_key:  # TODO: USE OpenKeyEx instead for subkeys...
            service_type = winreg.QueryValueEx(dev_key, "Service")[
                0
            ]  # TODO: ENUMIFY __name's
            if service_type == usb_service_type:
                name = winreg.QueryValueEx(dev_key, "FriendlyName")[0]
                desc = winreg.QueryValueEx(dev_key, "DeviceDesc")[0]
                usb_devices.append({name: desc})

    return usb_devices



from pathlib import Path

__all__ = ["sequence_files"]


def sequence_files(frames_dir: Path) -> None:
    """

    :param frames_dir:
    :type frames_dir:
    """
    for i, f in enumerate(sorted(frames_dir.iterdir(), key=lambda x: x.name)):
        if f.is_file():
            suffix = {f.suffix}
            # suffix = ".jpg"
            f.rename(frames_dir / f"{i:05d}{suffix}")


if __name__ == "__main__":
    sequence_files(
        Path(r"G:/") / "Mit drev" / "upscaled" / "restored_imgs",
    )



from pathlib import Path
from typing import Iterable

import cv2
import numpy
from draugr.opencv_utilities.color_space.threshold import hsv_min_max_clip_mask
from draugr.opencv_utilities.windows.elements import add_trackbar
from draugr.opencv_utilities.windows.image import show_image

__all__ = ["interactive_hsv_color_picker"]


def interactive_hsv_color_picker(
    ps: Iterable, wait_time: int = 33, verbose: bool = False
) -> None:
    """

    :param ps:
    :type ps:
    :param wait_time:
    :type wait_time:
    :param verbose:
    :type verbose:
    """
    window_label = "image"

    show_image(numpy.zeros((600, 600)), window_label)

    hmin_label = "HMin"
    hmax_label = "HMax"
    smin_label = "SMin"
    smax_label = "SMax"
    vmin_label = "VMin"
    vmax_label = "VMax"

    add_trackbar(
        window_label, hmin_label, default=0, max_val=179
    )  # Hue is from 0-179 for Opencv
    add_trackbar(window_label, hmax_label, default=179, max_val=179)

    add_trackbar(window_label, smin_label, default=0)
    add_trackbar(window_label, smax_label, default=255)

    add_trackbar(window_label, vmin_label, default=0)
    add_trackbar(window_label, vmax_label, default=255)

    # Initialize to check if HSV min/max value changes
    h_min = s_min = v_min = h_max = s_max = v_max = 0
    ph_min = ps_min = pv_min = ph_max = ps_max = pv_max = 0

    for p in ps:
        if not p.exists():
            continue

        img = cv2.imread(str(p))
        output = img
        while 1:
            h_min = cv2.getTrackbarPos(hmin_label, window_label)
            h_max = cv2.getTrackbarPos(hmax_label, window_label)

            s_min = cv2.getTrackbarPos(smin_label, window_label)
            s_max = cv2.getTrackbarPos(smax_label, window_label)

            v_min = cv2.getTrackbarPos(vmin_label, window_label)
            v_max = cv2.getTrackbarPos(vmax_label, window_label)

            output = cv2.bitwise_and(
                img,
                img,
                mask=hsv_min_max_clip_mask(
                    img,
                    numpy.array([h_min, s_min, v_min]),
                    numpy.array([h_max, s_max, v_max]),
                ),
            )

            # Print if there is a change in HSV value
            if (
                (ph_min != h_min)
                | (ps_min != s_min)
                | (pv_min != v_min)
                | (ph_max != h_max)
                | (ps_max != s_max)
                | (pv_max != v_max)
            ):
                if verbose:
                    print(
                        f"(hMin = {h_min:d} , sMin = {s_min:d}, vMin = {v_min:d}), (hMax = {h_max:d} , sMax = {s_max:d}, vMax = {v_max:d})"
                    )

                # TODO :look at neater handling of min-maxing values
                th_min = min(h_min, h_max)
                ts_min = min(s_min, s_max)
                tv_min = min(v_min, v_max)
                h_max = max(h_min, h_max)
                s_max = max(s_min, s_max)
                v_max = max(v_min, v_max)
                h_min = th_min
                s_min = ts_min
                v_min = tv_min

                cv2.setTrackbarPos(hmin_label, window_label, h_min)
                cv2.setTrackbarPos(hmax_label, window_label, h_max)
                cv2.setTrackbarPos(smin_label, window_label, s_min)
                cv2.setTrackbarPos(smax_label, window_label, s_max)
                cv2.setTrackbarPos(vmin_label, window_label, v_min)
                cv2.setTrackbarPos(vmax_label, window_label, v_max)

                ph_min = h_min
                ps_min = s_min
                pv_min = v_min
                ph_max = h_max
                ps_max = s_max
                pv_max = v_max

            if show_image(output, window_label, wait=wait_time):
                break

    cv2.destroyAllWindows()


if __name__ == "__main__":
    pss = (
        Path.home()
        / "ProjectsWin/AiBitbucket/Internal/OptikosPrime/exclude/new_images/adam_plus0p5/215asd.jpg",
        Path(r"C:\Users\Christian\OneDrive\Billeder\buh\7BIsT.png"),
        Path(r"C:\Users\Christian\OneDrive\Billeder\Portraits\thomas.jpg"),
    )
    if any(p.exists() for p in pss):
        interactive_hsv_color_picker(iter(pss))
    else:
        print(f"{pss} does not exist")
